/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
  * © 2016 OX Software GmbH, Germany. info@open-xchange.com
 *
 * @author Ingo Schmidt-Rosbiegal <ingo.schmidt-rosbiegal@open-xchange.com>
 */

/* eslint no-new: 0 */

define('io.ox/office/portal/mainhelper', [
    'io.ox/core/commons',
    'io.ox/office/tk/io',
    'io.ox/office/tk/utils/driveutils',
    'io.ox/office/portal/portalutils',
    'io.ox/office/portal/collection/documentcollection',
    'io.ox/office/portal/collection/templatecollection',
    'io.ox/office/portal/view/documentview',
    'io.ox/office/portal/view/templateview',
    'io.ox/office/portal/view/toolbarview',
    'io.ox/office/portal/view/tooltip',
    'io.ox/office/baseframework/app/guidedtour',
    'io.ox/office/editframework/app/fileactions',
    'settings!io.ox/office',
    'gettext!io.ox/office/portal/main',
    'less!io.ox/office/portal/style'
], function (Commons, IO, DriveUtils, Utils, DocumentCollection, TemplateCollection, DocumentView, TemplateView, ToolbarView, ToolTip, GuidedTour, FileActions, Settings, gt) {

    'use strict';

    // static class MainHelper ================================================

    var // the exported MainHelper class
        MainHelper = {};

    // methods ----------------------------------------------------------------

    MainHelper.createLauncher = function (appBaseName, appTitle, appOptions) {

        var // the module name of the portal application
            PORTAL_MODULE_NAME = Utils.getPortalModuleName(appBaseName),
            // the module name of the edit application
            EDIT_MODULE_NAME = Utils.getEditorModuleName(appBaseName),
            // the application instance
            app = ox.ui.createApp({ name: PORTAL_MODULE_NAME, title: appTitle, trackingId: 'io.ox/' + appBaseName + '-portal', appBaseName: appBaseName }),
            // RequireJS path of the application welcome tour
            appTourPath = Utils.getStringOption(appOptions, 'appTourPath', '');

        // load fileActions for the current App to use necessary action links for recent list/template context menus
        FileActions.create(appBaseName);

        app.setLauncher(function (options) {

            // create main window object
            var portalWindow = ox.ui.createWindow({
                name: PORTAL_MODULE_NAME,
                title: appTitle,
                chromeless: true
            });

            // the bootstrap of the OX Documents app
            function initOXDocuments() {

                // prepare app skeleton. Maybe this could be the AppView in the future.
                var rootNode = portalWindow.nodes.main,
                    windowBodyNode = portalWindow.nodes.body,
                    recentsNode = $('<div>').addClass('office-portal-recents').append($('<h2>').text(gt('Recent Documents'))),
                    recentsList = $('<div>').addClass('office-portal-recents-list f6-target'),
                    templatesNode = $('<div>').addClass('office-portal-templates').append($('<h2>').text(gt('New from Template'))),
                    // the document collection for the recent documents list
                    documentCollection = new DocumentCollection(),
                    // the template collection for the listed templates
                    templateCollection = new TemplateCollection();

                portalWindow.setTitle(appTitle);
                portalWindow.on('show', function () {
                    portalWindow.setTitle(appTitle);
                });

                // prefetch application source code
                IO.prefetchModuleSource(EDIT_MODULE_NAME);

                rootNode.empty().append(recentsNode.append(recentsList), templatesNode);

                // create application views
                new ToolbarView({ el: windowBodyNode, model: app });
                var documentView = new DocumentView({ collection: documentCollection, el: recentsList, model: app }),
                    templateView = new TemplateView({ collection: templateCollection, el: templatesNode, model: app });

                if (DriveUtils.isGuest()) {
                    // guests & anonymous users
                    documentCollection.setData([]);
                    templateCollection.setData([]);
                } else {
                    // try to get recent files and templates
                    portalWindow.busy();
                    IO.sendRequest(IO.FILTER_MODULE_NAME, { action: 'gettemplateandrecentlist', type: appBaseName })
                    .always(function () {
                        portalWindow.idle();
                    }).done(function (data) {
                        // populate the document and template collection with data from server
                        documentCollection.setData(data.recents || []);
                        templateCollection.setData(data.templates || []);
                        // render views with fresh data
                        documentView.render(appBaseName);
                        templateView.render(appBaseName);

                        // to init the bootstrap tooltips in portal
                        $('[data-toggle="tooltip"]').tooltip();

                        // start the welcome tour for this application once
                        GuidedTour.runAppTour(PORTAL_MODULE_NAME, { once: true, auto: true });
                    }).fail(appFail);
                }
            }

            // yells if something went wrong...
            function appFail(result) {
                require(['io.ox/core/capabilities', 'io.ox/core/notifications']).done(function (Capabilities, Notifications) {
                    var errorMsg = '';
                    if (!Capabilities.has('rt')) {
                        errorMsg += gt('A server component is not working. Please contact the server administrator.');
                    } else {
                        errorMsg += (result && result.error) ? result.error + ':' : '';
                        errorMsg += gt('Application may not work as expected until this problem is solved.');
                    }
                    Notifications.yell('error', errorMsg);
                });
                portalWindow.nodes.main.children().hide();
                portalWindow.nodes.body.children().hide();
            }

            // refreshes Documents app only on visibility
            function refresh() {
                if (!portalWindow.state.visible) { return; }
                initOXDocuments();
            }

            // save settings to the app object
            app.settings = Settings;

            portalWindow.addClass('io-ox-office-portal-main');

            // set it as the application window
            app.setWindow(portalWindow);

            // show portal window
            portalWindow.show();

            // Reinit app on portal window show. This is needed for this use case:
            // User opens a document from the recents list -> loses app focus,
            // and close the document -> gained focus to this app without an app 'resume' event).
            portalWindow.on('show', initOXDocuments);

            // refresh Documents portal on OX global refresh event
            ox.on('refresh^', refresh);

            // init OX Documents in folder support addition success.
            Commons.addFolderSupport(app, null, 'infostore', options.folder).always(initOXDocuments).fail(appFail);
        });

        // register as application tour (will create an entry in the settings menu to start the tour)
        if (appTourPath && !_.device('smartphone')) {
            GuidedTour.registerAppTour(PORTAL_MODULE_NAME, appTourPath);
        }

        return { getApp: app.getInstance };
    };

    // exports ================================================================

    return MainHelper;

});
